#!/usr/bin/env python3

__all__ = ['main']

import os
import sys
import argparse

def main(argv=None):
    if argv is None:
        argv = sys.argv[1:]
    parser = argparse.ArgumentParser('Grep')
    parser.add_argument('keys')
    parser.add_argument('type', nargs='?', help="Context information")
    parser.add_argument('-n', '--num', type=int)
    parser.add_argument('-w', '--whole')
    parser.add_argument('-k', '--keysep', default=',')
    parser.add_argument('-t', '--typesep', default=':')
    parser.add_argument('-d', '--debug', action='store_true')
    parser.add_argument('-o', '--only', action='store_true')

    args = parser.parse_args(argv)

    if args.whole:
        queries = [sys.stdin.read()]
    else:
        queries = sys.stdin.read().splitlines()

    if os.path.exists(args.keys):
        keys = open(args.keys).read().splitlines()
    else:
        keys = [a.strip() for a in args.keys.split(args.keysep)]

    if len(keys) == 1:
        keys.append('other')

    if args.type:
        dict = {key: f"{args.type}{args.typesep}{key}" for key in keys}
    else:
        dict = {key: key for key in keys}

    from wnix import Grep
    grep = Grep(queries, dict)
    if args.debug:
        print('dict:\n', dict, '\n', '='*42, file=sys.stderr)
        print('grep:\n', grep, '\n', '='*42, file=sys.stderr)
    output = grep.csv(n=args.num)
    print(output)

if __name__ == '__main__':
    main(sys.argv[1:])



#!/usr/bin/env python

import os

def test_grep_shell():

    cmd = "echo baseball basketball rugby soccer | fmt -1 | Grep 'european,american' sports"
    out = os.popen(cmd).read().splitlines()

    assert out == [
        'american:baseball',
        'american:basketball',
        'european:rugby',
        'european:soccer',
    ]



__all__ = [
    'Grep',
    'grep',
    'greps',
]

import assure
from .types import Object, List, Dict

SIMILARITY_DEFINITIONS = [
    'a @ b',
    'a @ b(a)',
    'a @ b(b)',
    'a(a) @ b',
    'a(a) @ b(a)',
    'a(a) @ b(b)',
    'a(b) @ b',
    'a(b) @ b(a)',
    'a(b) @ b(b)',
]


def argsort_reverse(a):
    import numpy as np
    return np.argsort(-a, axis=-1)

def promote(o):
    if isinstance(o, (list, tuple, set)):
        return List(o)
    elif isinstance(o, dict):
        return Dict(o)
    elif isinstance(o, str):
        return List([o])
    elif isinstance(o, bytes):
        return List([o.decode()])
    elif isinstance(o, (List, Dict)):
        return o
    else:
        raise TypeError(o)

class Grep:

    @classmethod
    def all(cls, queries, keys, *, n=None):
        """ Evaluate possible definitions of grep interactively
            by varying the definition of similarity """
        return [cls(queries, keys, n=n, sims=sims) for sims in cls.sims()]

    @classmethod
    def sims(cls):
        return SIMILARITY_DEFINITIONS

    def __init__(self, queries, keys, *, n=1, sims='a @ b'):
        q = promote(queries)
        k = promote(keys)
        if sims not in SIMILARITY_DEFINITIONS:
            raise ValueError(f"sims must be one of: {self.sims()}")
        s = (lambda a,b: eval(sims))(q,k)
        i = argsort_reverse(s)
        g = 0*s
        g.iloc[:, :] = s.columns.values[i]
        g.columns = list(range(1, len(g.columns)+1))
        g.columns.name = sims
        g = g.iloc[:, 0:n]

        # attributes
        self.queries = queries
        self.keys = keys
        self.q = q
        self.k = k
        self.s = s
        self.i = i
        self.g = g

    def __repr__(self):
        return repr(self.g)

    def __getitem__(self, n):
        return self.g[n]

    def csv(self, n=1):
        import pandas as pd
        g = self.g
        g = g.iloc[:, 0:n]
        q = g.index.tolist()
        k = g.to_csv(index=False, header=False)
        d = pd.DataFrame(q, index=k.splitlines())
        o = d.to_csv(index=True,header=False,sep=':').strip()
        return o

    def tolist(self):
        return self.g.values.ravel().tolist()

    def intersect(self, s):
        return set(self.tolist()) & set(s)

grep = Grep
greps = Grep.all



#!/usr/bin/env python3

__all__ = ['main']

import os
import sys

def main(argv=None):
    import argparse
    if argv is None:
        argv = sys.argv[1:]
    parser = argparse.ArgumentParser('What')
    parser.add_argument('file')
    args = parser.parse_args(argv)

    if args.file and os.path.exists(args.file):
        keys = open(args.file).read().splitlines()
    else:
        keys = sys.argv[1:]

    import embd
    import wnix
    q = embd.stdin_to_tensor()
    ranks = wnix.what(q, keys)
    output = '\n'.join(ranks)
    print(output)

if __name__ == '__main__':
    main(sys.argv[1:])



#!/usr/bin/env python3

__all__ = [
    'Object',
    'List',
    'Dict',
]

import embd
import is_instance

def series(o, space=None):
    import pandas as pd
    if space is None:
        space = embd
    if isinstance(o, pd.Series):
        return o
    if isinstance(o, str):
        return pd.Series(space.think(o), name=o)
    raise TypeError

def frame(o, space=None):
    import numpy as np
    import pandas as pd
    if space is None:
        space = embd
    if isinstance(o, pd.DataFrame):
        return o
    if isinstance(o, pd.Series):
        return pd.DataFrame(o)
    if is_instance(o, str):
        o = [o]
    if is_instance(o, list[str]):
        return pd.DataFrame(space.think(o).T, columns=o)
    if is_instance(o, dict[str, str]):
        return pd.DataFrame(space.think(list(o.values())).T, columns=list(o.keys()))
    if is_instance(o, np.ndarray):
        return pd.DataFrame(o.T, columns=list(range(len(o))))
    if is_instance(o, list[np.ndarray]):
        return pd.DataFrame(np.stack(o).T, columns=list(range(len(o))))
    if is_instance(o, dict[int, np.ndarray]):
        return pd.DataFrame(np.stack(list(o.values())).T, columns=list(o.keys()))
    raise TypeError(o)

# ======================

class O:
    def __init__(self, o):
        self.o = o

class D(O):
    def __init__(self, o):
        self.d = o
        super().__init__(o)

    def __getitem__(self, i):
        return self.d[i]

    def __len__(self):
        return len(self.d)

    def __iter__(self):
        return iter(self.d)

    def __repr__(self):
        return repr(self.d)

    def keys(self):
        return list(self.d.keys())

    def values(self):
        return list(self.d.values())

    def items(self):
        return list(self.d.items())

class L(O):
    def __init__(self, o):
        self.l = o
        super().__init__(o)

    def __getitem__(self, i):
        return self.l[i]

    def __len__(self):
        return len(self.l)

    def __iter__(self):
        return iter(self.l)

    def __add__(self, other):
        import itertools
        return L([''.join(p) for p in itertools.product(self, other)])

    def __repr__(self):
        return repr(self.l)

class F(O):
    def __init__(self, o):
        self.f = frame(o)
        super().__init__(o)

    def names(self):
        return self.f.columns

    def centerofmass(self):
        # center of mass: different from subtracting mean from self!
        return self.f.mean(axis=1)

    def dot(self, other):
        return self.f.T @ other.f

    def sims(self, other):
        return self.dot(self.at(self), other.at(other))

    def at(self, other):
        return F(self.f.subtract(other.centerofmass(), axis=0))

    def centered(self):
        return F(self.at(self))

    def centered_at(self, other):
        return self.at(other)

    def __matmul__(self, other):
        return self.dot(other)

    def __call__(self, other=None):
        if other is None: other = self
        return self.at(other)

    def __repr__(self):
        return repr(self.f)

    def __getitem__(self, i):
        return self.f[i]

    def __eq__(self, other):
        import numpy as np
        return np.allclose(self.f, other.f)

    def __sub__(self, other):
        return self.f - other.f

class S(O):
    def __init__(self, o):
        self.s = series(o)
        super().__init__(o)

    def name(self):
        return self.s.name

    def __sub__(self, other):
        return self.s - other.s

    def __repr__(self):
        return repr(self.s)


class Object(S):

    def __init__(self, o):
        super().__init__(o)

    def __repr__(self):
        return repr(self.s)

class List(F, L):

    def __init__(self, o):
        super().__init__(list(o))

    def __getitem__(self, i):
        if isinstance(i, int):
            return self.l[i]
        elif is_instance(i, str | list[str]):
            return self.f[i]
        else:
            raise IndexError(i)

    def __iter__(self):
        return iter(self.o)

class Dict(F, D):

    def __init__(self, o):
        super().__init__(dict(o))

    def __getitem__(self, i):
        if is_instance(i, str | list[str]):
            return self.f[i]
        else:
            raise IndexError(i)




"""
    sensory system
    turns hard-to-embed data types (pdfs, images, etc)
    into easy-to-embed data types (collections of tokens)
"""
from drivers import pdf
from drivers import html
from drivers import image



#!/usr/bin/env python3

__all__ = ['what']

# TODO: this module is a legacy implementation and will soon be replaced
# with a call to soft Grep with different arguments once we can make this work:
#
# from .grep import Grep
# def what(query, keys):
#     return Grep(query, keys, n=1)

import wnix
import assure
import numpy as np
from .types import List

def norm(A, axis=-1, keepdims=True):
    return np.linalg.norm(A, axis=axis, keepdims=keepdims)

def unit(A, axis=-1, keepdims=True):
    return A / norm(A, axis=axis, keepdims=keepdims)

def similarities(q, K):
    Q = q[None, :]
    K = unit(K)
    import scipy.special
    sims = scipy.special.softmax(Q @ K.T, axis=-1)
    return sims

def what(query, keys):
    q = assure.vector(query)
    K = List(keys).f.values.T
    s = similarities(q, K)
    I = np.argsort(s.ravel())[::-1]
    return np.array(keys)[I].tolist()
